/*
 * Copyright 2007 Geir Hedemark
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.testaco.test;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.MessageFormat;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import junit.framework.AssertionFailedError;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dbunit.Assertion;
import org.dbunit.DatabaseUnitException;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.CompositeDataSet;
import org.dbunit.dataset.DataSetException;
import org.dbunit.dataset.ITable;
import org.dbunit.dataset.NoSuchTableException;
import org.dbunit.dataset.filter.DefaultColumnFilter;
import org.dbunit.dataset.xml.FlatXmlWriter;
import org.dbunit.dataset.xml.XmlDataSet;
import org.dbunit.operation.DatabaseOperation;
import org.testaco.test.e.DataSetColumnCountException;
import org.testaco.test.e.DataSetRowCountException;
import org.testaco.test.e.DataSetSchemaException;
import org.testaco.test.e.ExtraColumnInContentException;
import org.testaco.test.e.ExtraTableInContentException;
import org.testaco.test.e.MissingColumnInContentException;
import org.testaco.test.e.MissingTableInContentException;


public class DataSetHelper {

	private static Log logger = LogFactory.getLog(DataSetHelper.class);
	
	public static void loadDataSet(final String databaseKey,
			final String content) throws Exception {
		IDatabaseConnection idbConn = DatabaseApplicationDetails.getInstance().getDbUnitConnection(databaseKey);
		clearDatabase(idbConn);

		if (content != null) {
			DatabaseOperation.CLEAN_INSERT.execute(idbConn, new XmlDataSet(
					new InputStreamReader(new FileInputStream(new File(
							content)), "utf-8")));
		}

	}

	private static void clearDatabase(IDatabaseConnection idbConn) throws SQLException {
		Connection conn = idbConn.getConnection();
		DatabaseMetaData md = conn.getMetaData();
		ResultSet rs = md.getTables(null, null, "%", null);
		while (rs.next()) {
			String tableName = rs.getString("TABLE_NAME");
			//TODO: Externalize sql for deletion
			conn.createStatement().execute(MessageFormat.format("delete from {0};", tableName));
			conn.createStatement().execute(
					//TODO: Externalize sql for reset of autogenerated keys
					MessageFormat.format("alter table {0} auto_increment=1;",
							tableName));
		}
	}
	
	protected static void checkDataset(
			final String databaseKey, final String resultsetfilename)
	throws Exception {
		IDatabaseConnection idbConn = DatabaseApplicationDetails.getInstance().getDbUnitConnection(databaseKey);
		ComparisonResult result = new ComparisonResult();
		result.content = idbConn.createDataSet();
		result.databaseKey = databaseKey;
		try { // For debugging/tracing operation
			try { // Functional - for rewriting exceptions
				compareDatasets(databaseKey, resultsetfilename, result);
			} catch (AssertionFailedError e) {
				if (e.getMessage()!=null && e.getMessage().contains("row count")) {
					DataSetRowCountException e2 = new DataSetRowCountException(result, e);
					throw e2;
				} else if (e.getMessage()!=null && e.getMessage().contains("column count")) {
					DataSetColumnCountException e2 = new DataSetColumnCountException(result, e);
					throw e2;
				} else {
					throw e;
				}
			}
		} catch (Exception e) {
			try {
				debugContents(result);
			} catch (Exception e2) {
				System.out.println("Oops while printing datasets! " + e2);
			}
			throw e;
		}
	}

	private static void debugContents(ComparisonResult result) throws IOException, DataSetException {
		System.out.println("Contents: ");
		FlatXmlWriter wr = new FlatXmlWriter(System.out);
		wr.write(result.content);
		System.out.println("Expected: ");
		wr = new FlatXmlWriter(System.out);
		wr.write(result.expected);
	}

	private static final Pattern countMatcher = Pattern.compile("expected:\\<(.*?)\\> but was:\\<(.*?)\\>",Pattern.MULTILINE | Pattern.DOTALL | Pattern.UNIX_LINES);

	private static void compareDatasets(final String databaseKey, final String resultsetfilename, ComparisonResult result) throws DataSetException, FileNotFoundException, DatabaseUnitException, ExtraTableInContentException, MissingTableInContentException {
		File resultSetFile = new File(resultsetfilename);
		result.expected = new CompositeDataSet(new XmlDataSet(
				new FileInputStream(resultSetFile)),
				new XmlDataSet(new FileInputStream(new File(
						DatabaseApplicationDetails.getInstance().getEmptyDataSetName(databaseKey)))));
		String[] contentNames = result.content.getTableNames();
		String[] expectedNames = result.expected.getTableNames();
		ITable expectedTable;
		ITable databaseTable;
		result.resultSetFilename = resultSetFile.getAbsolutePath();
//		result.resultSetFilename = resultsetfilename;
		for (int i = 0; i < contentNames.length; i++) {
			result.tableName = contentNames[i];
			Set<String> ignoredColumns = DatabaseApplicationDetails.getInstance().getIgnored(databaseKey, result.tableName);
			if (ignoredColumns==null) {
				ignoredColumns = new HashSet<String>();
			}
			
			try {
				expectedTable = result.expected.getTable(result.tableName);
			} catch (NoSuchTableException e) {
				ComparisonResult compResult = new ComparisonResult(result);
				compResult.databaseKey = databaseKey;
				compResult.tableName = result.tableName;
				throw new ExtraTableInContentException(compResult, e);
			}
			
			try {
				databaseTable = result.content.getTable(result.tableName);
			} catch (NoSuchTableException e) {
				ComparisonResult compResult = new ComparisonResult(result);
				compResult.databaseKey = databaseKey;
				compResult.tableName = result.tableName;
				throw new MissingTableInContentException(compResult, e);
			}
			Assertion.assertEquals(DefaultColumnFilter
					.excludedColumnsTable(expectedTable, ignoredColumns.toArray(new String[]{})),
					DefaultColumnFilter.excludedColumnsTable(databaseTable,
							ignoredColumns.toArray(new String[]{})));
		}
		Set<String> missingTables = new TreeSet<String>();
		for (int i = 0; i < expectedNames.length; i++) {
			String string = expectedNames[i];
			boolean found = false;
			if (string == null) {
				continue;
			}
			for (int j = 0; j < contentNames.length; j++) {
				if (string.equals(contentNames[j])) {
					found = true;
				}
			}
			if (!found) {
				missingTables.add(string);
			}
		}
		for (String missingTable : missingTables) {
			result.tableName = missingTable;
			Set<String> ignoredColumns = DatabaseApplicationDetails.getInstance().getIgnored(databaseKey, result.tableName);
			if (ignoredColumns==null) {
				ignoredColumns = new HashSet<String>();
			}
			//TODO: Throw missing table exceptions
			try {
				expectedTable = result.expected.getTable(result.tableName);
			} catch (NoSuchTableException e) {
				ComparisonResult compResult = new ComparisonResult(result);
				compResult.databaseKey = databaseKey;
				compResult.tableName = result.tableName;
				throw new ExtraTableInContentException(compResult, e);
			}
			try {
				databaseTable = result.content.getTable(result.tableName);
			} catch (NoSuchTableException e) {
				ComparisonResult compResult = new ComparisonResult(result);
				compResult.databaseKey = databaseKey;
				compResult.tableName = result.tableName;
				throw new MissingTableInContentException(compResult, e);
			}

			Assertion.assertEquals(DefaultColumnFilter
					.excludedColumnsTable(expectedTable, ignoredColumns.toArray(new String[]{})),
					DefaultColumnFilter.excludedColumnsTable(databaseTable,
							ignoredColumns.toArray(new String[]{})));
		}
	}

	public static void checkDatasetSchema(String databaseKey) throws Exception {
		String resultsetfilename = DatabaseApplicationDetails.getInstance().getEmptyDataSetName(databaseKey);
		IDatabaseConnection idbConn = DatabaseApplicationDetails.getInstance().getDbUnitConnection(databaseKey);
		ComparisonResult result = new ComparisonResult();
		result.content = idbConn.createDataSet();
		result.databaseKey = databaseKey;
		result.resultSetFilename = resultsetfilename;
		try { // For debugging/tracing operation
			try { // Functional - for rewriting exceptions
				compareDatasets(databaseKey, resultsetfilename, result);
			} catch (AssertionFailedError e) {
				//Four types of misfeatures:
				//Hopefully, this is schema related...
				if (e.getMessage()!=null && e.getMessage().contains("column count")) {
					int expected;
					int found;
					Matcher matcher = countMatcher.matcher(e.getMessage());
					if (matcher.find()) {
						expected = Integer.parseInt(matcher.group(1));
						found = Integer.parseInt(matcher.group(2));
						if (expected>found) {
							throw new MissingColumnInContentException(result, e);
						} else {
							throw new ExtraColumnInContentException(result, e);
						}
					} else {
						throw new IllegalArgumentException("Could not find contents of column count mismatch, original message was "+e.getMessage(), e);
					}
				} else {
					Exception e2 = new DataSetSchemaException(result, databaseKey, e);
					throw e2;
				}
			}
		} catch (Exception e) {
			try {
				debugContents(result);
			} catch (Exception e2) {
				System.out.println("Oops while printing datasets! " + e2);
			}
			throw e;
		}
	}

	public static void reloadDatabaseSchema(String databaseKey, String databaseSchemaDefinition) throws Exception {
		IDatabaseConnection idbConn = DatabaseApplicationDetails.getInstance().getDbUnitConnection(databaseKey);
		dropTables(idbConn); //In case someone didn't remember to drop all tables, ensures no gunk is left

		executeSql(idbConn, databaseSchemaDefinition);
	}

	private static void executeSql(IDatabaseConnection idbConn, String databaseSchemaDefinition) throws Exception {
		Connection conn = idbConn.getConnection();
		conn.createStatement().execute(databaseSchemaDefinition);
	}

	private static void dropTables(IDatabaseConnection idbConn) throws SQLException {
		Connection conn = idbConn.getConnection();
		DatabaseMetaData md = conn.getMetaData();
		//TODO: Externalize constants for this call - different for each database driver
		ResultSet rs = md.getTables(null, null, "%", null);
		if (logger.isDebugEnabled()) {
			logger.debug("Dropping all existing database tables for URL " + md.getURL());
		}
		while (rs.next()) {
			String tableName = rs.getString("TABLE_NAME");
			//TODO: Externalize sql for dropping tables
			Statement stmt = conn.createStatement();
			String query = MessageFormat.format("drop table {0};", tableName); 
			if (logger.isDebugEnabled()) {
				logger.debug("Dropping table " + md.getURL() + "." + tableName + " query is " + query);
			}
			stmt.execute(query);
		}
		if (logger.isDebugEnabled()) {
			logger.debug("All existing database tables dropped");
		}
	}

	
}
