/*
 * Copyright 2007 Geir Hedemark
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.testaco.test;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.DriverManager;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.ParserConfigurationException;


import org.dbunit.database.DatabaseConfig;
import org.dbunit.database.DatabaseConnection;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.ext.mysql.MySqlDataTypeFactory;
import org.testaco.test.configuration.DatabaseConfigurationDetails;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class DatabaseApplicationDetails {
	public static final String DEFAULT_CONFIGURATION_FILE = ".databasedefinitions.xml";
	@Deprecated
	public static final String CONFIGFILE = DEFAULT_CONFIGURATION_FILE;
	private static final String COLUMN_MODIFICATION_NAME_ATTRIBUTE = "name";
	private static final String TABLE_MODIFICATION_KEY_IGNORE = "ignore";
	private static final String TABLE_MODIFICATION_ELEMENT = "tableModification";
	private static final String DATABASE_DEFINITION_NAME_ATTRIBUTE = COLUMN_MODIFICATION_NAME_ATTRIBUTE;
	private static final String LOCALLY_DEFINED_DATABASE_DETAILS_ELEMENT_NAME = "localDatabaseDefinition";
	private Map<String, DatabaseConfigurationDetails> databaseConfigurations;

	private static Map<String, DatabaseConfigurationDetails> getSettings(String file) {
		Map<String, DatabaseConfigurationDetails> returnValue = new HashMap<String, DatabaseConfigurationDetails>();
		Document configXml = null;
		try {
			//TODO: reading wrong file

			configXml = javax.xml.parsers.DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
		} catch (SAXException e) {
			throw new RuntimeException("Could not read configuration file", e);
		} catch (IOException e) {
			throw new RuntimeException("Could not read configuration file", e);
		} catch (ParserConfigurationException e) {
			throw new RuntimeException("Could not read configuration file", e);
		}
		Node rootNode = configXml.getChildNodes().item(0);
		NodeList configElements = rootNode.getChildNodes();
		for (int i = 0; i < configElements.getLength(); i++) {
			Node configElement = configElements.item(i);
			if (configElement.getNodeName().endsWith(LOCALLY_DEFINED_DATABASE_DETAILS_ELEMENT_NAME)) {
				DatabaseConfigurationDetails databaseDetails = new DatabaseConfigurationDetails();
				String tableName = configElement.getAttributes().getNamedItem(DATABASE_DEFINITION_NAME_ATTRIBUTE).getTextContent();
				NodeList databaseConfigElements = configElement.getChildNodes();
				for (int j=0; j < databaseConfigElements.getLength(); j++) {
					Node databaseConfigElement = databaseConfigElements.item(j);
					if (databaseConfigElement.getNodeType()==Node.ELEMENT_NODE) {
						String localPart = databaseConfigElement.getNodeName();
						String value = databaseConfigElement.getTextContent();
						if (localPart.contains(":")) {
							localPart = databaseConfigElement.getNodeName().split(":")[1];
						}
						if (TABLE_MODIFICATION_ELEMENT.equals(localPart)) {
							NodeList columnModifications = databaseConfigElement.getChildNodes();
							for (int k=0; k < columnModifications.getLength(); k++) {
								Node columnModification = columnModifications.item(k);
								if (columnModification.getNodeType()==Node.ELEMENT_NODE) {
									String columnName = columnModification.getAttributes().getNamedItem(COLUMN_MODIFICATION_NAME_ATTRIBUTE).getTextContent();
									String modification = columnModification.getTextContent();
									if (modification!=null && TABLE_MODIFICATION_KEY_IGNORE.equals(modification.toLowerCase())) {
										databaseDetails.getIgnoredFields(tableName).add(columnName);
									} else {
										throw new IllegalArgumentException("Uknown modification type "+modification+
												" for database "+tableName+" and column "+columnName);
									}
								}
							}
						} else {
							String methodName = "set"+localPart.substring(0, 1).toUpperCase()+localPart.substring(1);
							Method setter;
							try {
								setter = databaseDetails.getClass().getMethod(methodName, new Class[]{String.class});
								setter.invoke(databaseDetails, new Object[] { value });
							} catch (NoSuchMethodException e) {
								throw new RuntimeException("Error while setting up database "+tableName+", could not find setter for "+localPart+" in DatabaseConfigurationDetails", e);
							} catch (IllegalAccessException e) {
								throw new RuntimeException("Error while setting up database "+tableName+", could not access setter for "+localPart+" in DatabaseConfigurationDetails", e);
							} catch (InvocationTargetException e) {
								throw new RuntimeException("Error while setting up database "+tableName+", could not invoke setter for "+localPart+" in DatabaseConfigurationDetails", e);
							} 
							returnValue.put(tableName, databaseDetails);
						}
					}
				}
			}
		}
		return returnValue;
	}

	public Set<String> getIgnored(String connectionTag, String tableName) {
		DatabaseConfigurationDetails configDetails = databaseConfigurations.get(connectionTag);
		if (configDetails==null) {
			throw new RuntimeException("Could not find configuration details for database "+connectionTag);
		}
		return Collections.unmodifiableSet(configDetails.getIgnoredFields(tableName));
	}
	
	public String getEmptyDataSetName(String connectionTag) {
		DatabaseConfigurationDetails configDetails = databaseConfigurations.get(connectionTag);
		if (configDetails==null) {
			throw new RuntimeException("Could not find configuration details for database "+connectionTag);
		}
		return configDetails.getEmptyDbUnitFileLocation();
	}

	public IDatabaseConnection getDbUnitConnection(final String connectionTagInput) throws Exception {
		String connectionTag = connectionTagInput;
		if (connectionTag==null) {
			connectionTag="default";
		}
		DatabaseConfigurationDetails configDetails = databaseConfigurations.get(connectionTag);
		if (configDetails==null) {
			throw new RuntimeException("Could not find configuration details for database "+connectionTag);
		}
		Class.forName(configDetails.getDriverClass());
		Connection jdbcConnection = DriverManager.getConnection( configDetails.getConnectionURI(),
				configDetails.getLoginName(),
				configDetails.getLoginPassword());
		IDatabaseConnection connection = new DatabaseConnection(
				jdbcConnection);
		if ("mysql".equals(configDetails.getDatabaseType())) {
			DatabaseConfig config = connection.getConfig();
			config.setProperty(DatabaseConfig.PROPERTY_DATATYPE_FACTORY,
					new MySqlDataTypeFactory());
			config.setProperty(" http://www.dbunit.org/features/datatypeWarning","false");
		}
		return connection;
	}
	public Set<String> getDefinedDatabaseKeys() {
		return databaseConfigurations.keySet();
	}

	private static DatabaseApplicationDetails instance;
	
	public static DatabaseApplicationDetails getInstance() {
		if (instance == null) {
			instance = new DatabaseApplicationDetails();
		}
		Map<String, DatabaseConfigurationDetails> tempvalue = new HashMap<String, DatabaseConfigurationDetails>();
		try {
			tempvalue = getSettings(DEFAULT_CONFIGURATION_FILE);
		} catch (Exception e) {
			//TODO log this, but dont fail in case we are running inside eclipse
		}
		instance.setDatabaseConfigurations(tempvalue);
		return instance;
	}

	public static DatabaseApplicationDetails getInstance(String configFile) {
		DatabaseApplicationDetails returnValue = new DatabaseApplicationDetails();
		Map<String, DatabaseConfigurationDetails> tempvalue = new HashMap<String, DatabaseConfigurationDetails>();
		tempvalue = getSettings(configFile);
		returnValue.setDatabaseConfigurations(tempvalue);
		
		return returnValue;
	}

	private void setDatabaseConfigurations(Map<String, DatabaseConfigurationDetails> tempvalue) {
		databaseConfigurations = tempvalue;
	}
	
	
}
