/*
 * Copyright 2007 Geir Hedemark
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.testaco.test;

import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.test.AbstractDependencyInjectionSpringContextTests;

/**
 * A test suite for handling database-aware test suites.
 * 
 * A test suite that encompasses a database (please see "Database refactoring" by Scott Ambler for more information)
 * needs to take into account changes in the database schema, and flag these changes as a refactored database,
 * not as a behavioral error caused by incorrect code.
 * <p>
 * It is assumed that the database schemas themselves are handled as a separate deliverable/artifact. The goal of a
 * database test suite is to
 *  - ensure that the owner of a code-based artifact using a database is made aware when changes are made to the database artifact
 *  - help support the development process when a database refactoring breaks behaviour. 
 * <p>
 * This test suite is the element of a database test suite which picks up on changes like these. All database
 * aware tests should be organized as part of this test suite. The tests contained within this test suite will
 * be run first in a test run, and will check the database content against a locally stored snapshot of the database without any content.
 * As part of these tests, the possibly refactored database schemas need to be reloaded. The user of this class needs to provide
 * a way of getting hold of a file containing the up-to-date database rebuild DDL commands. This is done using the
 * getDatabaseSchemaDefinition method
 * <p>
 * Any errors found during this phase of the test run will be reported first in the test resport. It is a very good idea to handle
 * these errors first before trying to deal with the other errors, as these are likely to be side effects of the changed database.
 * <p>
 * A db-unit xml based version of the database can be found locally. Please see the application setup file for details.
 * 
 * @see org.testaco.test.annotations.LoadDb
 * @see org.testaco.test.annotations.CheckDb
 * 
 * @author Geir Hedemark <geir@hedemark.net>
 *
 */
public abstract class DatabaseTestSuite extends AbstractDependencyInjectionSpringContextTests {
	public abstract String getDatabaseSchemaDefinition(String databaseKey) throws Exception;
	
	private Log logger = LogFactory.getLog(this.getClass());
	
	public void testDatabaseSchemas() throws Exception {
		//For all defined databases, check schema
		Set<String> databaseKeys = DatabaseApplicationDetails.getInstance().getDefinedDatabaseKeys();
		for (String databaseKey : databaseKeys) {
			if (logger.isDebugEnabled()) {
				logger.debug("Checking database schema for "+databaseKey);
			}
			DataSetHelper.reloadDatabaseSchema(databaseKey, getDatabaseSchemaDefinition(databaseKey));
			DataSetHelper.checkDatasetSchema(databaseKey);
		}
	}

	
}
