/*
 * Copyright 2007 Geir Hedemark
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.testaco.test;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;


import org.springframework.test.AbstractDependencyInjectionSpringContextTests;
import org.testaco.test.annotations.CheckDb;
import org.testaco.test.annotations.LoadDb;

/**
 * This class contains the test-time handling of the Testaco annotations.
 * <p/>
 * When the class is extended, all tests will be enhanced with code that tries to find LoadDb annotations before each test
 * starts. If any are found, the data sets mentioned in the annotation is loaded into the approporiate
 * database.
 * <p/>
 * After a test has completed, any CheckDb annotations are handled in a similar manner.
 * 
 * @see org.testaco.test.annotations.LoadDb
 * @see org.testaco.test.annotations.CheckDb
 *
 * @author geir
 *
 */
public abstract class SpringIntegrationTestCase extends
		AbstractDependencyInjectionSpringContextTests {
	
	/**
	 * Please override the onOnSetUp method to introduce setup code.
	 */
	@Override
	protected final void onSetUp() throws Exception {
		onOnSetUp();
		checkSetupAnnotations();
		super.onSetUp();
	}
	protected void onOnSetUp() {
		
	}
	
	/**
	 * Runs the bare test sequence.
	 * 
	 * This is an adaption of what junit does by default. The code is borrowed from the head of the
	 * junit source tree.
	 * 
	 * This method contains code from both Spring and Junit.
	 * @throws Throwable if any exception is thrown
	 */
	public void runBare() throws Throwable {
        // getName will return the name of the method being run
        if (isDisabledInThisEnvironment(getName())) {
                recordDisabled();
                logger.info("**** " + getClass().getName() + "." + getName() + " disabled in this environment: " +
                                "Total disabled tests=" + getDisabledTestCount());
                return;
        }

        Throwable exception= null;
		setUp();
		try {
			runTest();
		} catch (Throwable running) {
			exception= running;
		}
		finally {
			try {
				tearDown();
			} catch (Throwable tearingDown) {
				if (exception == null) exception= tearingDown;
			}
		}
		if (exception != null) throw exception;
	}
	/**
	 * Please override the onOnTearDown method to introduce tear down code.
	 */
	@Override
	protected final void onTearDown() throws Exception {
		onOnTearDown();
		
		checkTearDownAnnotations();
		super.onTearDown();
	}
	protected void onOnTearDown() {
		
	}
	
	/**
	 * TODO docs of setup annotations
	 * @throws Exception
	 */
	private void checkSetupAnnotations() throws Exception {
		//Unit under test
		String uut = this.getName(); 
		/*The "getName" method returns the name of the 
		 * method that is currently being tested in 
		 * junit 3.8.x. I hope that doesn't change. */
		Method uutMethod= null;
		try {
			uutMethod= getClass().getMethod(uut, (Class[])null);
		} catch (NoSuchMethodException e) {
			fail("Method \""+uut+"\" not found");
		}
		Annotation[] testAnnotations = uutMethod.getAnnotations();
		for (int i = 0; i < testAnnotations.length; i++) {
			Annotation annotation = testAnnotations[i];
			if (annotation instanceof LoadDb) {
				LoadDb loadDbAnnotation = (LoadDb) annotation;
				String databaseKey = loadDbAnnotation.database();
				String dataFileName = loadDbAnnotation.datafile();
				DataSetHelper.loadDataSet(databaseKey, dataFileName);
			}
		}
	}

	/**
	 * TODO docs of annotations
	 * @throws Exception
	 */
	private void checkTearDownAnnotations() throws Exception {
		String uut = this.getName();
		Method uutMethod= null;
		try {
			uutMethod= getClass().getMethod(uut, (Class[])null);
		} catch (NoSuchMethodException e) {
			fail("Method \""+uut+"\" not found");
		}
		Annotation[] testAnnotations = uutMethod.getAnnotations();
		for (int i = 0; i < testAnnotations.length; i++) {
			Annotation annotation = testAnnotations[i];
			if (annotation instanceof CheckDb) {
				CheckDb checkDbAnnotation = (CheckDb) annotation;
				String databaseKey = checkDbAnnotation.database();
				String dataFileName = checkDbAnnotation.datafile();
				DataSetHelper.checkDataset(databaseKey, dataFileName);
			}
		}
	}
}
